@@ -26,6 +26,8 @@
 
 #include <glib/gi18n-lib.h>
 #include <string.h>
+#include <time.h>
+
 #include "window.h"
 #include "class-group.h"
 #include "util.h"
@@ -64,7 +66,8 @@ static GHashTable *window_hash = NULL;
     ((window)->priv->is_fullscreen       << 8) |        \
     ((window)->priv->demands_attention   << 9) |        \
     ((window)->priv->is_urgent           << 10)|        \
-    ((window)->priv->is_above            << 11))
+    ((window)->priv->is_above            << 11)|        \
+    ((window)->priv->is_below            << 12))
 
 struct _WnckWindowPrivate
 {
@@ -118,6 +121,7 @@ struct _WnckWindowPrivate
   guint is_maximized_vert : 1;
   guint is_shaded : 1;
   guint is_above : 1;
+  guint is_below : 1;
   guint skip_pager : 1;
   guint skip_taskbar : 1;
   guint is_sticky : 1;
@@ -126,6 +130,8 @@ struct _WnckWindowPrivate
   guint demands_attention : 1;
   guint is_urgent : 1;
 
+  time_t needs_attention_time;
+
   /* _NET_WM_STATE_HIDDEN doesn't map directly into an
    * externally-visible state (it determines the WM_STATE
    * interpretation)
@@ -157,6 +163,7 @@ struct _WnckWindowPrivate
 };
 
 G_DEFINE_TYPE (WnckWindow, wnck_window, G_TYPE_OBJECT);
+#define WNCK_WINDOW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), WNCK_TYPE_WINDOW, WnckWindowPrivate))
 
 enum {
   NAME_CHANGED,
@@ -206,14 +213,86 @@ static guint signals[LAST_SIGNAL] = { 0 };
 static void
 wnck_window_init (WnckWindow *window)
 {
-  window->priv = g_new0 (WnckWindowPrivate, 1);
+  window->priv = WNCK_WINDOW_GET_PRIVATE (window);
 
+  window->priv->xwindow = None;
+  window->priv->name = NULL;
+  window->priv->app = NULL;
+  window->priv->class_group = NULL;
+  window->priv->group_leader = None;
+  window->priv->transient_for = None;
+  window->priv->icon_geometry.width = -1; /* invalid cached value */
   window->priv->name = NULL;
   window->priv->icon_name = NULL;
-  window->priv->workspace = ALL_WORKSPACES;
-  window->priv->icon_geometry.width = -1; /* invalid cached value */
+  window->priv->session_id = NULL;
+  window->priv->session_id_utf8 = NULL;
+  window->priv->pid = 0;
+  window->priv->workspace = -1;
+  window->priv->sort_order = G_MAXINT;
+
+  /* FIXME: should we have an invalid window type for this? */
+  window->priv->wintype = 0;
+
+  window->priv->icon = NULL;
+  window->priv->mini_icon = NULL;
 
   window->priv->icon_cache = _wnck_icon_cache_new ();
+
+  window->priv->actions = 0;
+
+  window->priv->x = 0;
+  window->priv->y = 0;
+  window->priv->width = 0;
+  window->priv->height = 0;
+
+  window->priv->left_frame = 0;
+  window->priv->right_frame = 0;
+  window->priv->top_frame = 0;
+  window->priv->bottom_frame = 0;
+
+  window->priv->startup_id = NULL;
+
+  window->priv->res_class = NULL;
+  window->priv->res_name = NULL;
+
+  window->priv->transient_for_root = FALSE;
+
+  window->priv->is_minimized = FALSE;
+  window->priv->is_maximized_horz = FALSE;
+  window->priv->is_maximized_vert = FALSE;
+  window->priv->is_shaded = FALSE;
+  window->priv->is_above = FALSE;
+  window->priv->is_below = FALSE;
+  window->priv->skip_pager = FALSE;
+  window->priv->skip_taskbar = FALSE;
+  window->priv->is_sticky = FALSE;
+  window->priv->is_hidden = FALSE;
+  window->priv->is_fullscreen = FALSE;
+  window->priv->demands_attention = FALSE;
+  window->priv->is_urgent = FALSE;
+
+  window->priv->needs_attention_time = 0;
+
+  window->priv->net_wm_state_hidden = FALSE;
+  window->priv->wm_state_iconic = FALSE;
+
+  window->priv->update_handler = 0;
+
+  window->priv->need_update_name = FALSE;
+  window->priv->need_update_state = FALSE;
+  window->priv->need_update_wm_state = FALSE;
+  window->priv->need_update_icon_name = FALSE;
+  window->priv->need_update_workspace = FALSE;
+  window->priv->need_update_actions = FALSE;
+  window->priv->need_update_wintype = FALSE;
+  window->priv->need_update_transient_for = FALSE;
+  window->priv->need_update_startup_id = FALSE;
+  window->priv->need_update_wmclass = FALSE;
+  window->priv->need_update_wmhints = FALSE;
+  window->priv->need_update_frame_extents = FALSE;
+
+  window->priv->need_emit_name_changed = FALSE;
+  window->priv->need_emit_icon_changed = FALSE;
 }
 
 static void
@@ -221,6 +300,8 @@ wnck_window_class_init (WnckWindowClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
+  g_type_class_add_private (klass, sizeof (WnckWindowPrivate));
+
   object_class->finalize = wnck_window_finalize;
 
   /**
@@ -247,8 +328,8 @@ wnck_window_class_init (WnckWindowClass *klass)
    *
    * Emitted when the state of @window changes. This can happen when @window is
    * (un)minimized, (un)maximized, (un)sticked, (un)shaded, (un)made above,
-   * (un)set fullscreen, when it needs attention, etc. See #WnckWindowState for
-   * the complete list of states that might have changed.
+   * (un)made below, (un)set fullscreen, when it needs attention, etc. See
+   * #WnckWindowState for the complete list of states that might have changed.
    */
   signals[STATE_CHANGED] =
     g_signal_new ("state_changed",
@@ -336,25 +417,40 @@ wnck_window_finalize (GObject *object)
 
   unqueue_update (window);
 
+  if (window->priv->app)
+    g_object_unref (G_OBJECT (window->priv->app));
+  window->priv->app = NULL;
+
+  if (window->priv->class_group)
+    g_object_unref (G_OBJECT (window->priv->class_group));
+  window->priv->class_group = NULL;
+
+  g_free (window->priv->name);
+  window->priv->name = NULL;
+  g_free (window->priv->icon_name);
+  window->priv->icon_name = NULL;
+  g_free (window->priv->session_id);
+  window->priv->session_id = NULL;
+  g_free (window->priv->session_id_utf8);
+  window->priv->session_id_utf8 = NULL;
+
   if (window->priv->icon)
     g_object_unref (G_OBJECT (window->priv->icon));
+  window->priv->icon = NULL;
 
   if (window->priv->mini_icon)
     g_object_unref (G_OBJECT (window->priv->mini_icon));
+  window->priv->mini_icon = NULL;
   
   _wnck_icon_cache_free (window->priv->icon_cache);
+  window->priv->icon_cache = NULL;
 
-  g_free (window->priv->name);
-  g_free (window->priv->icon_name);
-  g_free (window->priv->session_id);
   g_free (window->priv->startup_id);
+  window->priv->startup_id = NULL;
   g_free (window->priv->res_class);
+  window->priv->res_class = NULL;
   g_free (window->priv->res_name);
-  
-  if (window->priv->app)
-    g_object_unref (G_OBJECT (window->priv->app));
-
-  g_free (window->priv);
+  window->priv->res_name = NULL;
 
   G_OBJECT_CLASS (wnck_window_parent_class)->finalize (object);
 }
@@ -587,6 +683,34 @@ wnck_window_get_icon_name (WnckWindow *window)
     return FALLBACK_NAME;
 }
 
+char *
+_wnck_window_get_name_for_display (WnckWindow *window,
+                                   gboolean    use_icon_name,
+                                   gboolean    use_state_decorations)
+{
+  const char *name;
+
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
+  
+  if (use_icon_name && wnck_window_has_icon_name (window))
+    name = wnck_window_get_icon_name (window);
+  else 
+    name = wnck_window_get_name (window);
+  
+  if (use_state_decorations)
+    {
+      if (window->priv->is_shaded)
+        return g_strdup_printf ("=%s=", name);
+      else if (window->priv->is_minimized)
+        return g_strdup_printf ("[%s]", name);
+      else
+        return g_strdup (name);
+    }
+  else
+    return g_strdup (name);
+}
+
+
 /**
  * wnck_window_get_application:
  * @window: a #WnckWindow.
@@ -902,15 +1026,23 @@ wnck_window_needs_attention (WnckWindow *window)
   return window->priv->demands_attention || window->priv->is_urgent;
 }
 
-/* Return whether one of the transients of @window needs attention */
-static gboolean
+time_t
+_wnck_window_get_needs_attention_time (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), 0);
+
+  return window->priv->needs_attention_time;
+}
+
+/* Return whether the transient of @window needs attention */
+static WnckWindow *
 transient_needs_attention (WnckWindow *window)
 {
   GList *windows;
   WnckWindow *transient;
   
   if (!WNCK_IS_WINDOW (window))
-    return FALSE;
+    return NULL;
 
   windows = wnck_screen_get_windows_stacked (window->priv->screen);
 
@@ -919,15 +1051,34 @@ transient_needs_attention (WnckWindow *window)
     {
       /* catch transient cycles */
       if (transient == window)
-        return FALSE;
+        return NULL;
 
       if (wnck_window_needs_attention (transient))
-        return TRUE;
+        return transient;
     }
 
   return FALSE;
 }
 
+time_t
+_wnck_window_or_transient_get_needs_attention_time (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), 0);
+
+  if (_wnck_window_get_needs_attention_time (window) == 0)
+    {
+      WnckWindow *transient;
+
+      transient = transient_needs_attention (window);
+      if (transient)
+        return _wnck_window_get_needs_attention_time (transient);
+      else
+        return 0;
+    }
+  else
+    return _wnck_window_get_needs_attention_time (window);
+}
+
 /**
  * wnck_window_or_transient_needs_attention:
  * @window: a #WnckWindow.
@@ -944,7 +1095,7 @@ gboolean
 wnck_window_or_transient_needs_attention (WnckWindow *window)
 {
   return wnck_window_needs_attention (window) || 
-         transient_needs_attention (window);
+         transient_needs_attention (window) != NULL;
 }
 
 /**
@@ -1082,6 +1233,27 @@ wnck_window_is_above                  (WnckWindow *window)
   return window->priv->is_above;
 }
 
+/**
+ * wnck_window_is_below:
+ * @window: a #WnckWindow.
+ *
+ * Returns whether @window is below other windows. This state may change
+ * anytime a #WnckWindow::state-changed signal gets emitted.
+ *
+ * See wnck_window_make_below() for more details on this state.
+ *
+ * Return value: %TRUE if @window is below other windows, %FALSE otherwise.
+ *
+ * Since: 2.20
+ **/
+gboolean
+wnck_window_is_below                  (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
+
+  return window->priv->is_below;
+}
+
 /**
  * wnck_window_is_skip_pager:
  * @window: a #WnckWindow.
@@ -1226,7 +1398,10 @@ wnck_window_is_sticky                 (WnckWindow *window)
  *
  * Closes @window.
  *
- * Since: 2.6 (@timestamp argument was missing in earlier versions)
+ * This function existed before 2.6, but the @timestamp argument was missing
+ * in earlier versions.
+ * 
+ * Since: 2.6
  **/
 void
 wnck_window_close (WnckWindow *window,
@@ -1457,6 +1632,47 @@ wnck_window_unmake_above (WnckWindow *window)
                       0);
 }
 
+/**
+ * wnck_window_make_below:
+ * @window: a #WnckWindow.
+ *
+ * Asks the window manager to put @window below most windows.
+ *
+ * Since: 2.20
+ **/
+void
+wnck_window_make_below (WnckWindow *window)
+{
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                      window->priv->xwindow,
+                      TRUE,
+                      _wnck_atom_get ("_NET_WM_STATE_BELOW"),
+                      0);
+}
+
+/**
+ * wnck_window_unmake_below:
+ * @window: a #WnckWindow.
+ *
+ * Asks the window manager to not put @window below most windows, and to
+ * put it again in the stack with other windows.
+ *
+ * Since: 2.20
+ **/
+void
+wnck_window_unmake_below (WnckWindow *window)
+{
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                      window->priv->xwindow,
+                      FALSE,
+                      _wnck_atom_get ("_NET_WM_STATE_BELOW"),
+                      0);
+}
+
 /**
  * wnck_window_stick:
  * @window: a #WnckWindow.
@@ -1499,7 +1715,7 @@ wnck_window_unstick                 (WnckWindow *window)
  * wnck_window_keyboard_move:
  * @window: a #WnckWindow.
  *
- * Asks the window manager to start moving @window via keyboard.
+ * Asks the window manager to start moving @window via the keyboard.
  **/
 void
 wnck_window_keyboard_move (WnckWindow *window)
@@ -1514,7 +1730,7 @@ wnck_window_keyboard_move (WnckWindow *window)
  * wnck_window_keyboard_size:
  * @window: a #WnckWindow.
  *
- * Asks the window manager to start resizing @window via keyboard.
+ * Asks the window manager to start resizing @window via the keyboard.
  **/
 void
 wnck_window_keyboard_size (WnckWindow *window)
@@ -1640,7 +1856,10 @@ wnck_window_unpin (WnckWindow *window)
  * decide to refuse the request (to not steal the focus if there is a more
  * recent user activity, for example).
  *
- * Since: 2.10 (@timestamp argument was missing in earlier versions)
+ * This function existed before 2.10, but the @timestamp argument was missing
+ * in earlier versions.
+ * 
+ * Since: 2.10
  **/
 void
 wnck_window_activate (WnckWindow *window,
@@ -1746,7 +1965,10 @@ find_last_transient_for (GList *windows,
  * the most recently active window among @window and its transients.
  * This is probably best implemented on the window manager side.
  * 
- * Since: 2.10 (@timestamp argument was missing in earlier versions)
+ * This function existed before 2.10, but the @timestamp argument was missing
+ * in earlier versions.
+ * 
+ * Since: 2.10
  **/
 void
 wnck_window_activate_transient (WnckWindow *window,
@@ -2392,11 +2614,16 @@ update_state (WnckWindow *window)
   
   if (reread_net_wm_state)
     {
+      gboolean demanded_attention;
+
+      demanded_attention = window->priv->demands_attention;
+
       window->priv->is_maximized_horz = FALSE;
       window->priv->is_maximized_vert = FALSE;
       window->priv->is_sticky = FALSE;
       window->priv->is_shaded = FALSE;
       window->priv->is_above = FALSE;
+      window->priv->is_below = FALSE;
       window->priv->skip_taskbar = FALSE;
       window->priv->skip_pager = FALSE;
       window->priv->net_wm_state_hidden = FALSE;
@@ -2424,6 +2651,8 @@ update_state (WnckWindow *window)
             window->priv->is_shaded = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_ABOVE"))
             window->priv->is_above = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_BELOW"))
+            window->priv->is_below = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_FULLSCREEN"))
             window->priv->is_fullscreen = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_TASKBAR"))
@@ -2436,6 +2665,14 @@ update_state (WnckWindow *window)
           ++i;
         }
 
+      if (window->priv->demands_attention != demanded_attention)
+        {
+          if (window->priv->demands_attention)
+            time (&window->priv->needs_attention_time);
+          else if (!window->priv->is_urgent)
+            window->priv->needs_attention_time = 0;
+        }
+
       g_free (atoms);
     }
 
@@ -2838,9 +3075,16 @@ update_wmhints (WnckWindow *window)
           window->priv->group_leader = hints->window_group;
 
       if (hints->flags & XUrgencyHint)
-        window->priv->is_urgent = TRUE;
+        {
+          window->priv->is_urgent = TRUE;
+          time (&window->priv->needs_attention_time);
+        }
       else
-        window->priv->is_urgent = FALSE;
+        {
+          window->priv->is_urgent = FALSE;
+          if (!window->priv->demands_attention)
+            window->priv->needs_attention_time = 0;
+        }
 
       XFree (hints);
     }
