@@ -9,6 +9,10 @@ GType wnck_pager_display_mode_get_type (void);
 /* enumerations from "screen.h" */
 GType wnck_motion_direction_get_type (void);
 #define WNCK_TYPE_MOTION_DIRECTION (wnck_motion_direction_get_type())
+GType _wncklayoutorientation___wnck_layout_orientation_get_type (void);
+#define WNCK_TYPE___WNCK_LAYOUT_ORIENTATION (_wncklayoutorientation___wnck_layout_orientation_get_type())
+GType _wncklayoutcorner___wnck_layout_corner_get_type (void);
+#define WNCK_TYPE___WNCK_LAYOUT_CORNER (_wncklayoutcorner___wnck_layout_corner_get_type())
 /* enumerations from "tasklist.h" */
 GType wnck_tasklist_grouping_type_get_type (void);
 #define WNCK_TYPE_TASKLIST_GROUPING_TYPE (wnck_tasklist_grouping_type_get_type())
